/*
 * Copyright SpringSource Inc 2007,2009.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.springsource.util.parser.manifest;

import com.springsource.util.parser.manifest.internal.SourceContext;

/**
 * <p>
 * Encapsulates a problem discovered during manifest parsing.
 * 
 * <strong>Concurrent Semantics</strong><br />
 * 
 * This class is thread safe.
 * 
 * @author Andy Clement
 */
public class ManifestProblem {

	private ManifestProblemKind problemKind;
	private SourceContext context;
	private int startoffset;
	private int endoffset;
	private String[] inserts;

	public ManifestProblem(ManifestProblemKind problem, SourceContext context, int startoffset, int endoffset, String... inserts) {
		this.problemKind = problem;
		this.context = context;
		this.startoffset = startoffset;
		this.endoffset = endoffset;
		this.inserts = inserts;
	}

	@Override
	public String toString() {
		return problemKind.format(context.getLine(startoffset), context.getColumn(startoffset), inserts);
	}

	/**
	 * Returns a multi line message that includes the manifest extract that gave
	 * rise to the problem.
	 * 
	 * @return a multi line string
	 */
	public String toStringWithContext() {
		int relevantLine = context.getLine(startoffset);
		String relevantLineData = context.getLineAsString(relevantLine);
		String theMessage = problemKind.format(context.getLine(startoffset), context.getColumn(startoffset), inserts);
		StringBuilder beautifulMessage = new StringBuilder();
		beautifulMessage.append(relevantLineData);
		int scol = context.getColumn(startoffset);
		int ecol = context.getColumn(endoffset);
		int where = 0;
		for (where = 0; where < scol; where++) {
			beautifulMessage.append(' ');
		}
		beautifulMessage.append('^');
		if (ecol != scol) {
			where++;
			for (; where < ecol; where++) {
				beautifulMessage.append(' ');
			}
			beautifulMessage.append('^');
		}
		beautifulMessage.append('\n');
		beautifulMessage.append(theMessage);
		return beautifulMessage.toString();
	}

	public ManifestProblemKind getKind() {
		return problemKind;
	}

	public int getInsertCount() {
		return inserts.length;
	}

	public String getInsert(int insertIndex) {
		return inserts[insertIndex];
	}

	public int getStartLine() {
		return context.getLine(startoffset);
	}

	public int getEndLine() {
		return context.getLine(endoffset);
	}

	public int getStartColumn() {
		return context.getColumn(startoffset);
	}

	public int getEndColumn() {
		return context.getColumn(endoffset);
	}

	public int getStartOffset() {
		return startoffset;
	}

	public int getEndOffset() {
		return endoffset;
	}

}
