/*
 * Copyright SpringSource Inc 2007,2009.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.springsource.util.parser.manifest.internal;

import java.util.ArrayList;
import java.util.List;

import com.springsource.util.parser.manifest.ManifestProblem;

/**
 * 
 * <strong>Concurrent Semantics</strong><br/>
 * 
 * This class is thread safe.
 * 
 * @author Andy Clement
 */
public class SimpleTokenStream implements TokenStream {

	private int tokenStreamPosition = 0;
	private List<Token> tokens = new ArrayList<Token>();
	private int tokenStreamLen;
	private List<ManifestProblem> problems = new ArrayList<ManifestProblem>();
	private SourceContext sourceContext;

	SimpleTokenStream() {
		tokens.clear();
		tokenStreamPosition = 0;
	}

	@Override
	public String toString() {
		return toFormattedString(true);
	}

	public int getCount() {
		return tokens.size();
	}

	public int getPosition() {
		return tokenStreamPosition;
	}

	public Token next() {
		if (tokenStreamPosition >= tokenStreamLen) {
			return null;
		} else {
			return tokens.get(tokenStreamPosition++);
		}
	}

	public void setPosition(int newPosition) {
		tokenStreamPosition = newPosition;
	}

	public Token peek() {
		if (tokenStreamPosition >= tokens.size()) {
			return tokens.get(tokens.size() - 1);
		}
		return tokens.get(tokenStreamPosition);
	}

	public Token peekLast() {
		return tokens.get(tokens.size() - 1);
	}

	public Token peek(int offset) {
		// TODO [later] negative offset check >0
		if ((tokenStreamPosition + offset) >= tokenStreamLen) {
			return null;
		} else {
			return tokens.get(tokenStreamPosition + offset);
		}
	}

	public String toFormattedString() {
		return toFormattedString(true);
	}

	public String toFormattedString(boolean includePositionsInOutput) {
		StringBuilder sb = new StringBuilder();
		sb.append("TokenStream:#").append(tokens.size()).append(" tokens:");
		sb.append("[");
		int i = 0;
		for (Token token : tokens) {
			if (i > 0) {
				sb.append(",");
			}
			if (includePositionsInOutput) {
				sb.append(token.toString());
			} else {
				sb.append(token.value());

			}
			i++;
		}
		sb.append("]");
		return sb.toString();
	}

	void addToken(Token token) {
		tokens.add(token);
	}

	public boolean containsProblems() {
		return !problems.isEmpty();
	}

	public List<ManifestProblem> getProblems() {
		return problems;
	}

	public SourceContext getSourceContext() {
		return sourceContext;
	}

	public boolean hasMore() {
		return tokenStreamPosition < tokens.size();
	}

	public void recordProblem(ManifestProblem manifestProblem) {
		problems.add(manifestProblem);
	}

	public void setSourceContext(SourceContext sourceContext) {
		this.sourceContext = sourceContext;
	}

	/**
	 * Called by the lexer to indicate that population of the tokenStream is now
	 * complete.
	 */
	public void lexComplete() {
		tokenStreamLen = tokens.size();
	}

}
